/*******************************************************************************
 * Copyright (c) 2008 Intel Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Rodolfo Kohn - initial API and implementation
 *    Marcelo Lorenzatti, Miriam Estela - Management of multiple configurations
      and filters .
 *******************************************************************************
 Product name: Plugin Extension for CIM  
 Creation date: 1-Nov-2007
 Description: 
 This file checks and loads the parameters of a plugin in an data array provided 
 by command line args

 Modification History
 02/20/2008      Marcelo Lorenzati, Miriam Estela    Fixed to read more than one 
                 functionality per plugin
 02/29/2008      Marcelo Lorenzati    Check stages for PEC_PluginFunctionalitySet
                 tingData and PEC_RelatedIndication.
********************************************************************************/
#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <cmci.h>
#include <native.h>

#include "libpec_configurationmanagement.h"

#define BUFFER_SIZE 256
#define INDICATION_ID_LEN 20
#define PLUGIN_NAME_LEN 256
#define OPT_AND_ARGS_LEN 256
#define ARRAY_SIZE 300
#define TARGET_OS_LINUX 36
#define SW_ELEM_STATE_EXECUTABLE 2
#define PLUGIN_VERSION "1.0.0"
#define NAMESPACE "root/cimv2"
#define PLUGIN_CLASS                 "PEC_Plugin"
#define PLUGIN_FUNCTIONALITY_CLASS   "PEC_PluginFunctionalitySettingData"
#define PLUGIN_FUNC_INDICATION_CLASS "PEC_RelatedIndication"
#define RELATED_PROPERTY_CLASS       "PEC_RelatedProperty"
#define PLUGIN_FUNC_ASSOCIATION      "CIM_ElementSettingData"
#define FUNC_IND_ASSOCIATION         "PEC_FunctionalityIndicationSupport"
#define FUNCTIONALITY_DESC_LEN 256
#define OPTIONS_ARGS_LEN 256
#define PLUGIN_LEN 64
#define SW_ELEM_ID_LEN 5
#define IND_ID_LEN 20
#define PROB_CAUSE_LEN 256
#define SRC_CLASS_LEN 256
#define INSTANCE_ID_LEN 64
#define INSTANCE_ID_PREFIX "PEC:"
#define INSTRUMENTED_CLASS_LEN 256
#define INSTRUMENTED_PROPERTY_LEN 256

#define DEFAULT_CIM_HOST "localhost"
#define DEFAULT_USER "root"
#define DEFAULT_PASSWORD "pec123"

typedef enum {FALSE = 0, TRUE} bool_t;
typedef enum {ALL = 0, GET = 1, INDICATION = 2} functionalityMode_t;
typedef struct
{    char pluginName[PLUGIN_NAME_LEN + 1];
    char indicationIdentifier[INDICATION_ID_LEN + 1];
    int  frequency;
    char optAndArgs[OPT_AND_ARGS_LEN + 1];
} pluginEventID_t;

pluginEventID_t PECPLUGIN_pluginEventIDArray[ARRAY_SIZE];
int pluginEventIDArrayLen = 0;


void showObj(CMPIObjectPath *objectpath);

void getInstanceData(CMPIInstance *instance,
                     pluginEventID_t *PECPLUGIN_pluginEventIDArray,
                     int idx);

int createAssociationFunctionalityInd(CMCIClient *cc,
                                      const char *namespace,
                                      const char *class,
                                      CMPIObjectPath *objectpath1, 
                                      CMPIObjectPath *objectpath2);

int createAssociationPluginFunctionality(CMCIClient *cc,
                                         const char *namespace,
                                         const char *class,
                                         CMPIObjectPath *objectpath1, 
                                         CMPIObjectPath *objectpath2, 
                                         int def);

int createRelatedProperty(CMCIClient *cc,
                          const char *namespace,
                          const char *class,
                          const char *instrumentedClass,
                          const char *instrumentedProperty,
                          CMPIObjectPath **objectpath);

int checkRelatedIndication(CMCIClient *cc,
                           const char *namespace,
                           const char *class,
                           const char *indicationIdentifier,
                           CMPIObjectPath **objectpath);

int createRelatedIndication(CMCIClient *cc,
                            const char *namespace,
                            const char *class,
                            const char *indicationIdentifier,
                            const char *defaultProbableCause,
                            uint16_t frequency,
                            const char *sourceClass,
                            CMPIObjectPath **objectpath);

int createPluginInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *softwareElementID,
                         const char *name,
                         const char *version,
                         bool_t concurrencyAllowed,
                         CMPIObjectPath **objectpath);

void setPluginValues(CMPIObjectPath *objectpath,
                     const char *softwareElementID,
                     const char *name,
                     const char *version);

int checkPluginFunctionalityInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *instanceID,
                         CMPIObjectPath **objectpath);

int createPluginFunctionalityInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *instanceID,
                         const char *optionsAndArguments,
                         const char *functionalityDescription,
                         functionalityMode_t mode,
                         CMPIObjectPath **objectpath);

/*This function print to stdout the path, namespace and classname of the passed objectpath*/
void showObj(CMPIObjectPath *objectpath)
{
    CMPIString * objectpathname = objectpath->ft->toString(objectpath, NULL);
    CMPIString * namespace = objectpath->ft->getNameSpace(objectpath, NULL);
    CMPIString * classname = objectpath->ft->getClassName(objectpath, NULL);

    if (objectpathname && objectpathname->hdl)
    {
        printf("objectpath=%s\n", (char *)objectpathname->hdl);
    }
    if (namespace && namespace->hdl)
    {
        printf("namespace=%s\n", (char *)namespace->hdl);
    }
    if (classname && classname->hdl)
    {
        printf("classname=%s\n", (char *)classname->hdl);
    }
}

/*This function retrieves  the related data of an Instance*/
void getInstanceData(CMPIInstance *instance,
                     pluginEventID_t *PECPLUGIN_pluginEventIDArray,
                     int idx)
{
    pluginEventID_t pluginData;
    CMPIObjectPath * opInstance = instance->ft->getObjectPath(instance, NULL);
    CMPIString * opInstanceName = opInstance->ft->toString(opInstance, NULL);
    CMPIString * iNameSpace = opInstance->ft->getNameSpace(opInstance, NULL);
    CMPIString * iClassName = opInstance->ft->getClassName(opInstance, NULL);
    int countkeys = opInstance->ft->getKeyCount(opInstance, NULL);
    int countProperties = instance->ft->getPropertyCount(instance, NULL);
    int i;

    if (opInstanceName && opInstanceName->hdl)
    {
        printf("Objectpath=%s\n", (char *)opInstanceName->hdl);
    }
    if (iNameSpace && iNameSpace->hdl)
    {
        printf("Namespace=%s\n", (char *)iNameSpace->hdl);
    }
    if (iClassName && iClassName->hdl)
    {
        printf("Classname=%s\n", (char *)iClassName->hdl);
    }
    if (countkeys)
    {
        printf("Keys:\n");
        CMPIData data;
        CMPIString * keyname;
        for (i = 0; i < countkeys; i++)
        {
            data = opInstance->ft->getKeyAt(opInstance, i, &keyname, NULL);
            char *ptr = NULL;
            ptr = (char *) value2Chars(data.type, &data.value);
            if (strcmp((char *)keyname->hdl, "Name") == 0)
            {
                strcpy(pluginData.pluginName, ptr);
            }
            printf("\t%s=%s\n", (char *)keyname->hdl, (ptr));
            free(ptr);
            if (keyname) CMRelease(keyname);
        }
    }
    else { printf("No keys!!\n"); }
    if (countProperties)
    {
        printf("Properties:\n");
        for (i = 0; i < countProperties; i++)
        {
            CMPIString * propertyname;
            CMPIData propData = instance->ft->getPropertyAt(instance, i, &propertyname, NULL);
            if (strcmp((char *)propertyname->hdl, "DefaultIndicationIdentifier") == 0)
            {
                char *ptr = NULL;
                ptr = (char *) value2Chars(propData.type, &propData.value);
                strcpy(pluginData.indicationIdentifier, ptr);
                free(ptr);
            }
            if (strcmp((char *)propertyname->hdl, "DefaultFrequency") == 0)
            {
                pluginData.frequency = propData.value.uint16;
            }
            if (strcmp((char *)propertyname->hdl, "DefaultOptionsAndArguments") == 0)
            {
                char *ptr = NULL;
                ptr = (char *) value2Chars(propData.type, &propData.value);
                strcpy(pluginData.optAndArgs, ptr);
                free(ptr);
            }
            CMRelease(propertyname);
        }
    }
    else
    {
        printf("No properties!!\n");
    }
    memcpy(&PECPLUGIN_pluginEventIDArray[idx], &pluginData, sizeof(pluginData));
    if (iClassName) CMRelease(iClassName);
    if (iNameSpace) CMRelease(iNameSpace);
    if (opInstanceName) CMRelease(opInstanceName);
    if (opInstance) CMRelease(opInstance);
}

/*This function creates an association for the passed functionality*/
int createAssociationFunctionalityInd(CMCIClient *cc,
                                      const char *namespace,
                                      const char *class,
                                      CMPIObjectPath *objectpath1, 
                                      CMPIObjectPath *objectpath2)
{
    CMPIObjectPath *objectpath;
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    int res = 0;
    
    showObj(objectpath1);
    showObj(objectpath2);
    objectpath = newCMPIObjectPath(namespace, class, NULL);
    value.ref = objectpath1;
    CMAddKey(objectpath, "Functionality", &value, CMPI_ref); 
    value.ref = objectpath2;
    CMAddKey(objectpath, "Indication", &value, CMPI_ref); 
    instance = newCMPIInstance(objectpath, NULL);
    objectpath_r = cc->ft->createInstance(cc, objectpath, instance, &status);
    if (status.rc || !objectpath_r) 
    {
        printf("Error in createAssociationFunctionalityInd: createInstance () rc = %d, msg = %s\n", 
                status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    if (objectpath) CMRelease(objectpath);
    return res;
}

/*This function creates an association for the PEC_PluginFunctionalitySettingData */
int createAssociationPluginFunctionality(CMCIClient *cc,
                                         const char *namespace,
                                         const char *class,
                                         CMPIObjectPath *objectpath1, 
                                         CMPIObjectPath *objectpath2, 
                                         int def)
{
    CMPIObjectPath *objectpath;
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    int res = 0;
    
    showObj(objectpath1);
    showObj(objectpath2);
    objectpath = newCMPIObjectPath(namespace, class, NULL);
    value.ref = objectpath1;
    CMAddKey(objectpath, "ManagedElement", &value, CMPI_ref); 
    value.ref = objectpath2;
    CMAddKey(objectpath, "SettingData", &value, CMPI_ref); 
    instance = newCMPIInstance(objectpath, NULL);
    if (!instance) 
    {
        printf("Failed creating instance\n");
    }
    value.uint16 = def;
    CMSetProperty(instance, "IsDefault", &value, CMPI_uint16);
    objectpath_r = cc->ft->createInstance(cc, objectpath, instance, &status);
    if (status.rc || !objectpath_r) 
    {
        printf("Error in createAssociationPluginFunctionality: createInstance () rc = %d, msg = %s\n", 
                status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    if (objectpath) CMRelease(objectpath);
    return res;
}


int createRelatedProperty(CMCIClient *cc,
                          const char *namespace,
                          const char *class,
                          const char *instrumentedClass,
                          const char *instrumentedProperty,
                          CMPIObjectPath **objectpath)
{
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    int res = 0;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    CMAddKey(*objectpath, "InstrumentedClassName", instrumentedClass, CMPI_chars);
    CMAddKey(*objectpath, "InstrumentedPropertyName", instrumentedProperty, CMPI_chars);
    instance = newCMPIInstance(*objectpath, NULL);
    objectpath_r = cc->ft->createInstance(cc, *objectpath, instance, &status);
    if (status.rc) 
    {
        printf("Error in createRelatedProperty: createInstance () rc = %d, msg = %s\n", 
                status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    return res;
}


int checkRelatedIndication(CMCIClient *cc,
                           const char *namespace,
                           const char *class,
                           const char *indicationIdentifier,
                           CMPIObjectPath **objectpath)
{
    CMPIInstance *instance;
    CMPIStatus status;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    CMAddKey(*objectpath, "IndicationIdentifier", indicationIdentifier, CMPI_chars);
    instance = cc->ft->getInstance(cc, *objectpath, 0, NULL, &status);
    if (status.rc == 0)
    {
        printf("Related Indication Instance already exist.\n");
    }
    else if (status.rc != 6)
    {
        printf("Error on checkRelatedIndication.\n");
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);

    return (status.rc != 6);
}

/*This function creates the Indication instance of a functionality*/
int createRelatedIndication(CMCIClient *cc,
                            const char *namespace,
                            const char *class,
                            const char *indicationIdentifier,
                            const char *defaultProbableCause,
                            uint16_t frequency,
                            const char *sourceClass,
                            CMPIObjectPath **objectpath)
{
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    int res = 0;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    CMAddKey(*objectpath, "IndicationIdentifier", indicationIdentifier, CMPI_chars);
    instance = newCMPIInstance(*objectpath, NULL);
    value.uint16 = frequency;
    CMSetProperty(instance, "Frequency", &value, CMPI_uint16);
    CMSetProperty(instance, "DefaultProbableCause", defaultProbableCause, CMPI_chars);
    CMSetProperty(instance, "SourceClass", sourceClass, CMPI_chars);
    objectpath_r = cc->ft->createInstance(cc, *objectpath, instance, &status);
    if (status.rc) 
    {
        printf("Error in createRelatedIndications: createInstance () rc = %d, msg = %s\n", 
               status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    return res;    
}

/*This function creates the Instance of the plugin*/
int createPluginInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *softwareElementID,
                         const char *name,
                         const char *version,
                         bool_t concurrencyAllowed,
                         CMPIObjectPath **objectpath)
{
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    CMPIData data;
    int res = 0;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    setPluginValues(*objectpath, softwareElementID, name, version);
    instance = newCMPIInstance(*objectpath, NULL);
    value.boolean = concurrencyAllowed;
    CMSetProperty(instance, "ConcurrencyAllowed", &value, CMPI_boolean);
    objectpath_r = cc->ft->createInstance(cc, *objectpath, instance, &status);
    if (status.rc == 11)
    {
        data = (*objectpath)->ft->getKey(*objectpath, "Name", &status);
        printf("PluginInstance %s already exists:\n", (char *)((data.value.string)->hdl));
    }
    else if (status.rc)
    {
        printf("Error in createPluginInstance: createInstance () rc = %d, msg = %s\n", 
               status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    return res;
}


void setPluginValues(CMPIObjectPath *objectpath,
                     const char *softwareElementID,
                     const char *name,
                     const char *version)
{
    CMPIValue value;
    uint16_t targetOS = TARGET_OS_LINUX;
    uint16_t swElemState = SW_ELEM_STATE_EXECUTABLE;

    value.uint16 = targetOS;
    CMAddKey(objectpath, "TargetOperatingSystem", &value, CMPI_uint16); 
    value.uint16 = swElemState;
    CMAddKey(objectpath, "SoftwareElementState", &value, CMPI_uint16); 
    CMAddKey(objectpath, "SoftwareElementID", softwareElementID, CMPI_chars);
    CMAddKey(objectpath, "Name", name, CMPI_chars);
    CMAddKey(objectpath, "Version", version, CMPI_chars);
}

int checkPluginFunctionalityInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *instanceID,
                         CMPIObjectPath **objectpath)
{
    CMPIInstance *instance;
    CMPIStatus status;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    CMAddKey(*objectpath, "InstanceID", instanceID, CMPI_chars);
    instance = cc->ft->getInstance(cc, *objectpath, 0, NULL, &status);
    if (status.rc == 0)
    {
        printf("Plugin Functionality Instance already exist.\n");
    }
    else if (status.rc != 6)
    {
        printf("Error on checkPluginFunctionalityinstance.\n");
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);

    return (status.rc != 6);
}

/*This function creates the functionalityinstance of a PLUGIN*/
int createPluginFunctionalityInstance(CMCIClient *cc,
                         const char *namespace,
                         const char *class,
                         const char *instanceID,
                         const char *optionsAndArguments,
                         const char *functionalityDescription,
                         functionalityMode_t mode,
                         CMPIObjectPath **objectpath)
{
    CMPIObjectPath *objectpath_r;
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIValue value;
    int res = 0;

    *objectpath = newCMPIObjectPath(namespace, class, NULL);
    CMAddKey(*objectpath, "InstanceID", instanceID, CMPI_chars); 
    instance = newCMPIInstance(*objectpath, NULL);
    CMSetProperty(instance, "OptionsAndArguments", optionsAndArguments, CMPI_chars);
    CMSetProperty(instance, "Description", functionalityDescription, CMPI_chars);
    value.uint16 = mode;
    CMSetProperty(instance, "Mode", &value, CMPI_uint16); 
    objectpath_r = cc->ft->createInstance(cc, *objectpath, instance, &status);
    if (status.rc) 
    {
        printf("Error in createPluginFunctionalityInstance: createInstance () rc = %d, msg = %s\n", 
                status.rc, (status.msg) ? (char *)(status.msg->hdl) : NULL);
        res = -1;
    }
    if (status.msg)   CMRelease(status.msg);
    if (instance)     CMRelease(instance);
    if (objectpath_r) CMRelease(objectpath_r);
    return res;
}


int main(int argc, char **argv)
{
    CMCIClient *cc;
    CMPIStatus status;
    char *cim_host;
    char *cim_host_passwd;
    char *cim_host_userid;
    char plugin[PLUGIN_LEN + 1];
    char optionsAndArgs[OPTIONS_ARGS_LEN + 1];
    char functionalityDesc[FUNCTIONALITY_DESC_LEN + 1];
    char softwareElementID[SW_ELEM_ID_LEN + 1];
    functionalityMode_t mode;
    char indicationIdentifier[IND_ID_LEN + 1];
    char probableCause[PROB_CAUSE_LEN + 1];
    char sourceClass[SRC_CLASS_LEN + 1];
    char instanceID[INSTANCE_ID_LEN + 1];
    char instrumentedClass[INSTRUMENTED_CLASS_LEN + 1];
    char instrumentedProperty[INSTRUMENTED_PROPERTY_LEN + 1];
    uint16_t freq;
    int res;
    int i, j;
    CMPIObjectPath *objectpath1;
    CMPIObjectPath *objectpath2;
    CMPIObjectPath *objectpath3;
    CMPIObjectPath *objectpath4;

    if (argc < 7)
    {
        printf("Too few arguments\n");
        return 1;
    }

    strncpy(plugin, argv[1], PLUGIN_LEN);
    strncpy(optionsAndArgs, argv[2], OPTIONS_ARGS_LEN);
    strncpy(functionalityDesc, argv[3], FUNCTIONALITY_DESC_LEN);
    strcpy(softwareElementID, "00000");

    for(i = (strlen(argv[4]) - 1), j = (SW_ELEM_ID_LEN - 1); i >= 0; i--, j--)
    {
        softwareElementID[j] = argv[4][i];
    }
    if (strcmp(argv[5], "Indication") == 0) mode = INDICATION;
    else if (strcmp(argv[5], "Get") == 0) mode = GET;
    else if (strcmp(argv[5], "All") == 0) mode = ALL;
    sprintf(instanceID, "%s%s", INSTANCE_ID_PREFIX, argv[6]);
    if (mode == INDICATION || mode == ALL)
    {
        if (argc < 11)
        {
            printf("Too few arguments\n");
            return 1;
        }
        strncpy(indicationIdentifier, argv[7], IND_ID_LEN);
        strncpy(probableCause, argv[8], PROB_CAUSE_LEN);
        freq = (uint16_t) atoi(argv[9]);
        strncpy(sourceClass, argv[10], SRC_CLASS_LEN);
    }

    pecConfigurationManagementInit();

    cim_host = pecConfigurationManagementGetValue("CIM_SERVER");
    if (cim_host == NULL) { cim_host = DEFAULT_CIM_HOST; }
    cim_host_userid = pecConfigurationManagementGetValue("USER");
    if (cim_host_userid == NULL) { cim_host_userid = DEFAULT_USER; }
    cim_host_passwd = pecConfigurationManagementGetValue("PASSWORD");
    if (cim_host_passwd == NULL) { cim_host_passwd = DEFAULT_PASSWORD; }
    cc = cmciConnect(cim_host, NULL, "5988", cim_host_userid, cim_host_passwd, NULL);

    if (!cc)
    {
        printf("Could not initialize connection.\n");
        return 1;
    }

    if (checkPluginFunctionalityInstance(cc, NAMESPACE, PLUGIN_FUNCTIONALITY_CLASS,
                                         instanceID,  &objectpath2))
    {
        if (cc) CMRelease(cc);
        return 2; 
    }

    if (checkRelatedIndication(cc, NAMESPACE, PLUGIN_FUNC_INDICATION_CLASS,
                               indicationIdentifier, &objectpath3))
    {
        if (cc) CMRelease(cc);
        return 3; 
    }

    res = createPluginInstance(cc, NAMESPACE, PLUGIN_CLASS, softwareElementID,
                               plugin, PLUGIN_VERSION, TRUE, &objectpath1);
    res = createPluginFunctionalityInstance(cc, NAMESPACE, PLUGIN_FUNCTIONALITY_CLASS,
                                            instanceID, optionsAndArgs,
                                            functionalityDesc, mode, &objectpath2); 
    res = createAssociationPluginFunctionality(cc, NAMESPACE, PLUGIN_FUNC_ASSOCIATION, 
                                               objectpath1, objectpath2, 1);
    if (mode == INDICATION || mode == ALL)
    {
        createRelatedIndication(cc, NAMESPACE, PLUGIN_FUNC_INDICATION_CLASS,
                                indicationIdentifier, probableCause, freq,
                                sourceClass, &objectpath3);
        createAssociationFunctionalityInd(cc, NAMESPACE, FUNC_IND_ASSOCIATION,
                                          objectpath2, objectpath3);
    }
    if (mode == GET || mode == ALL)
    {
        createRelatedProperty(cc, NAMESPACE, RELATED_PROPERTY_CLASS, 
                              instrumentedClass, instrumentedProperty, 
                              &objectpath4);
    }
    if (cc) CMRelease(cc);
    return 0;
}
